
var _Module = require('module');
var path = require('path');
var util = require('util');
var assert = require('assert').ok;
var debug = function(){};

var Script = process.binding('evals').NodeScript;
var runInThisContext = Script.runInThisContext;
var runInNewContext = Script.runInNewContext;
var NativeModule = { require: require, exists: function() { return false; } };

function Module(id, parent) {
  this.id = id;
  this.exports = {};
  this.parent = parent;

  this.filename = null;
  this.loaded = false;
  this.exited = false;
  this.children = [];
}

util.inherits(Module,_Module);

Module._extensions = _Module._extensions;
Module._pathCache = _Module._pathCache;
Module._cache = _Module._cache;
Module.wrapper = _Module.wrapper;
Module.wrap = _Module.wrap;

function statPath(path) {
  var fs = NativeModule.require('fs');
  try {
    return fs.statSync(path);
  } catch (ex) {}
  return false;
}
// check if the directory is a package.json dir
var packageCache = {};

function readPackage(requestPath) {
  if (packageCache.hasOwnProperty(requestPath)) {
    return packageCache[requestPath];
  }

  var fs = NativeModule.require('fs');
  try {
    var jsonPath = path.resolve(requestPath, 'package.json');
    var json = fs.readFileSync(jsonPath, 'utf8');
    var pkg = packageCache[requestPath] = JSON.parse(json);
    return pkg;
  } catch (e) {}

  return false;
}

function tryTopSCPackage(requestPath, exts) {
    var spl = requestPath.split('/');
    var topPath = spl.slice(0,spl.length-1).join('/');
    var pkg = readPackage(topPath);
    //console.log('topPath',topPath);
    if (!pkg) { return false; }
    var pkgdir = path.resolve(topPath,'packages',spl.slice(spl.length-1));
    var inpkg = readPackage(pkgdir);
    if (!inpkg) { return false; }
    var filename = path.resolve(pkgdir,(inpkg.directories || {}).lib,inpkg.main);
      return tryFile(filename) || tryExtensions(filename, exts) ||
             tryExtensions(path.resolve(filename, 'index'), exts) ||
             tryExtensions(path.resolve(filename, 'main'), exts);
}

function trySCPackage(requestFrom, request, exts) {
  var requestPath = path.resolve(requestFrom,request);
  var filename, 
      pkg = readPackage(requestPath);
  if (pkg) {
      filename = path.resolve(requestPath,(pkg.directories || {}).lib,pkg.main);
      return tryFile(filename) || tryExtensions(filename, exts) ||
      tryExtensions(path.resolve(filename, 'index'), exts) ||
      tryExtensions(path.resolve(filename, 'main'), exts);
  } else {
    var spl = request.split('/'), 
        i = spl.length-1, topPath, restPath, r;
    while (i >= 0) {
        topPath = path.resolve(requestFrom,spl.slice(0,i).join('/'));
        restPath = spl.slice(i,spl.length).join('/');        
        pkg = readPackage(topPath);
        if (pkg) {
          filename = path.resolve(topPath,(pkg.directories || {}).lib,restPath);
          r = tryFile(filename) || tryExtensions(filename, exts);
          if (r) { return r; }
          r = trySCPackage(path.resolve(topPath,'packages'),restPath,exts);
          if (r) { return r; }
        }
        --i;
    }
    return false;
  }
}

function tryPackage(requestPath, exts) {
  //var r = trySCPackage(requestPath, exts);
  //if (r) { return r; }
  
  var pkg = readPackage(requestPath);

  if (!pkg || !pkg.main) return false;

  var filename = path.resolve(requestPath, pkg.main);
  return tryFile(filename) || tryExtensions(filename, exts) ||
         tryExtensions(path.resolve(filename, 'index'), exts);
}

// check if the file exists and is not a directory
function tryFile(requestPath) {
  var fs = NativeModule.require('fs');
  var stats = statPath(requestPath);
  if (stats && !stats.isDirectory()) {
    return fs.realpathSync(requestPath, Module._realpathCache);
  }
  return false;
}

// given a path check a the file exists with any of the set extensions
function tryExtensions(p, exts) {
  for (var i = 0, EL = exts.length; i < EL; i++) {
    var filename = tryFile(p + exts[i]);

    if (filename) {
      return filename;
    }
  }
  return false;
}


Module._findPath = function(request, paths) {
  var fs = NativeModule.require('fs');
  var exts = Object.keys(Module._extensions);

  if (request.charAt(0) === '/') {
    paths = [''];
  }

  var trailingSlash = (request.slice(-1) === '/');

  var cacheKey = JSON.stringify({request: request, paths: paths});
  if (Module._pathCache[cacheKey]) {
    return Module._pathCache[cacheKey];
  }

  // For each path
  for (var i = 0, PL = paths.length; i < PL; i++) {
    var basePath = path.resolve(paths[i], request);
    var filename;

    if (!trailingSlash) {
      // try to join the request to the path
      filename = tryFile(basePath);

      if (!filename && !trailingSlash) {
        // try it with each of the extensions
        filename = tryExtensions(basePath, exts);
      }
    }

    if (!filename) {
      filename = tryPackage(basePath, exts);
    }
    if (!filename) {
      filename = trySCPackage(paths[i],request,exts);
    }

    if (!filename) {
      // try it with each of the extensions at "index"
      filename = tryExtensions(path.resolve(basePath, 'index'), exts);
    }

    if (filename) {
      Module._pathCache[cacheKey] = filename;
      return filename;
    }
  }
  return false;
};


// 'from' is the __dirname of the module.
Module._nodeModulePaths = function(from) {
  // guarantee that 'from' is absolute.
  from = path.resolve(from);

  // note: this approach *only* works when the path is guaranteed
  // to be absolute.  Doing a fully-edge-case-correct path.split
  // that works on both Windows and Posix is non-trivial.
  var splitRe = process.platform === 'win32' ? /[\/\\]/ : /\//;
  // yes, '/' works on both, but let's be a little canonical.
  var joiner = process.platform === 'win32' ? '\\' : '/';
  var paths = [];
  var parts = from.split(splitRe);

  for (var tip = parts.length - 1; tip >= 0; tip --) {
    // don't search in .../node_modules/node_modules
    if (parts[tip] !== 'node_modules') {
        var dir = parts.slice(0, tip + 1).concat('node_modules').join(joiner);
        paths.push(dir);
    }
    if (parts[tip] !== 'packages') {
        var dir = parts.slice(0, tip + 1).concat('packages').join(joiner);
        paths.push(dir);
    }
  }

  return paths;
}


Module._resolveLookupPaths = function(request, parent) {
  if (NativeModule.exists(request)) {
    return [request, []];
  }

  var start = request.substring(0, 2);
  if (start !== './' && start !== '..') {
    var paths = Module._paths;
    if (parent) {
      if (!parent.paths) parent.paths = [];
      paths = parent.paths.concat(paths);
    }
    return [request, paths];
  }

  // with --eval, parent.id is not set and parent.filename is null
  if (!parent || !parent.id || !parent.filename) {
    // make require('./path/to/foo') work - normally the path is taken
    // from realpath(__filename) but with eval there is no filename
    var mainPaths = ['.'].concat(Module._paths);
    mainPaths = Module._nodeModulePaths('.').concat(mainPaths);
    return [request, mainPaths];
  }

  // Is the parent an index module?
  // We can assume the parent has a valid extension,
  // as it already has been accepted as a module.
  var isIndex = /^index\.\w+?$/.test(path.basename(parent.filename));
  var parentIdPath = isIndex ? parent.id : path.dirname(parent.id);
  var id = path.resolve(parentIdPath, request);

  // make sure require('./path') and require('path') get distinct ids, even
  // when called from the toplevel js file
  if (parentIdPath === '.' && id.indexOf('/') === -1) {
    id = './' + id;
  }

  debug('RELATIVE: requested:' + request +
        ' set ID to: ' + id + ' from ' + parent.id);

  return [id, [path.dirname(parent.filename)]];
};

Module._load = function(request, parent, isMain) {
    
  if (parent) {
    debug('Module._load REQUEST  ' + (request) + ' parent: ' + parent.id);
  }

  var resolved = Module._resolveFilename(request, parent);
  var id = resolved[0];
  var filename = resolved[1];

  var cachedModule = Module._cache[filename];
  if (cachedModule) {
    return cachedModule.exports;
  }

  if (NativeModule.exists(id)) {
    // REPL is a special case, because it needs the real require.
    if (id == 'repl') {
      var replModule = new Module('repl');
      replModule._compile(NativeModule.getSource('repl'), 'repl.js');
      NativeModule._cache.repl = replModule;
      return replModule.exports;
    }

    debug('load native module ' + request);
    return NativeModule.require(id);
  }

  var module = new Module(id, parent);

  if (isMain) {
    process.mainModule = module;
    module.id = '.';
  }

  Module._cache[filename] = module;
  try {
    module.load(filename);
  } catch (err) {
    delete Module._cache[filename];
    throw err;
  }

  return module.exports;
};

Module._resolveFilename = function(request, parent) {
  if (NativeModule.exists(request)) {
    return [request, request];
  }

  var resolvedModule = Module._resolveLookupPaths(request, parent);
  var id = resolvedModule[0];
  var paths = resolvedModule[1];
  
  
  // look up the filename first, since that's the cache key.
  debug('looking for ' + JSON.stringify(id) +
        ' in ' + JSON.stringify(paths));

  var filename = Module._findPath(request, paths);
  if (!filename) {
      //console.log('--------------------');
    throw new Error("Cannot find module '" + request + "'");
  }
  id = filename;
  return [id, filename];
};



Module.prototype.load = function(filename) {
  debug('load ' + JSON.stringify(filename) +
        ' for module ' + JSON.stringify(this.id));

  assert(!this.loaded);
  this.filename = filename;
  this.paths = Module._nodeModulePaths(path.dirname(filename));

  var extension = path.extname(filename) || '.js';
  if (!Module._extensions[extension]) extension = '.js';
  Module._extensions[extension](this, filename);
  this.loaded = true;
};


// Returns exception if any
Module.prototype._compile = function(content, filename) {
  var self = this;
  // remove shebang
  content = content.replace(/^\#\!.*/, '');

  function require(path) {
    return Module._load(path, self);
  }

  require.resolve = function(request) {
    return Module._resolveFilename(request, self)[1];
  }
  require.paths = Module._paths;
  require.main = process.mainModule;
  // Enable support to add extra extension types
  require.extensions = Module._extensions;
  require.registerExtension = function() {
    throw new Error('require.registerExtension() removed. Use ' +
                    'require.extensions instead.');
  }
  require.cache = Module._cache;

  var dirname = path.dirname(filename);
  var define = makeDefine(require, self);

  if (Module._contextLoad) {
    if (self.id !== '.') {
      debug('load submodule');
      // not root module
      var sandbox = {};
      for (var k in global) {
        sandbox[k] = global[k];
      }
      sandbox.require = require;
      sandbox.exports = self.exports;
      sandbox.__filename = filename;
      sandbox.__dirname = dirname;
      sandbox.module = self;
      sandbox.global = sandbox;
      sandbox.root = root;
      sandbox.define = define;

      return runInNewContext(content, sandbox, filename, true);
    }

    debug('load root module');
    // root module
    global.require = require;
    global.exports = self.exports;
    global.__filename = filename;
    global.__dirname = dirname;
    global.module = self;
    global.define = define;

    return runInThisContext(content, filename, true);
  }

  // create wrapper function
  var wrapper = Module.wrap(content);

  var compiledWrapper = runInThisContext(wrapper, filename, true);
  if (filename === process.argv[1] && global.v8debug) {
    global.v8debug.Debug.setBreakPoint(compiledWrapper, 0, 0);
  }

  var args = [self.exports, require, self, filename, dirname, define];
  return compiledWrapper.apply(self.exports, args);
};

// AMD compatibility
function makeDefine(require, module) {
  var called = false;
  function define() {
    if (called) {
      throw new Error("define() may only be called once.");
    }
    called = true;

    // only care about the last argument
    var cb = arguments[ arguments.length - 1 ];

    // set exports immediately:
    // define({ foo: "bar" })
    if (typeof cb !== 'function') {
      module.exports = cb;
      return;
    }

    var ret = cb(require, module.exports, module);

    if (ret !== undefined) {
      // set exports with return statement
      // define(function () { return { foo: "bar" } })
      module.exports = ret;
    }
  }

  return define;
}

Module._scload = function(request, parent, isMain) {
    try {
        return _Module._load(request, parent, isMain);
    } catch (e) {
        if (e.message.indexOf('Cannot find module ') == 0) {
            return Module._load(request, parent, isMain);
        } else {
            throw e;
        }
    }
};

exports.SCModule = Module;